/*    
      glm.c
      Nate Robins, 1997, 2000
      nate@pobox.com, http://www.pobox.com/~nate
 
      Wavefront OBJ model file format reader/writer/manipulator.

      Includes routines for generating smooth normals with
      preservation of edges, welding redundant vertices & texture
      coordinate generation (spheremap and planar projections) + more.
  
*/
#pragma comment(linker, "/NODEFAULTLIB")



#include <windows.h>
#include <gl/gl.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "glm.h"

#define T(x) (model->triangles[(x)])

#define TRUE 1
#define FALSE 0 

/* _GLMnode: general purpose node */
typedef struct _GLMnode {
    int         index;
    int      averaged;
    struct _GLMnode* next;
} GLMnode;

char *_mystrdup(char *name) {
    int namelen = strlen(name);
    char *newstr = malloc(namelen + 1);
    strcpy(newstr, name);
    newstr[namelen] = 0;
    return(newstr);
}


/* glmMax: returns the maximum of two floats */
static float
glmMax(float a, float b) 
{
    if (b > a)
        return b;
    return a;
}

/* glmAbs: returns the absolute value of a float */
static float
glmAbs(float f)
{
    if (f < 0)
        return -f;
    return f;
}

/* glmDot: compute the dot product of two vectors
 *
 * u - array of 3 floats (float u[3])
 * v - array of 3 floats (float v[3])
 */
static float
glmDot(float* u, float* v)
{
    //assert(u); assert(v);
    
    return u[0]*v[0] + u[1]*v[1] + u[2]*v[2];
}

/* glmCross: compute the cross product of two vectors
 *
 * u - array of 3 floats (float u[3])
 * v - array of 3 floats (float v[3])
 * n - array of 3 floats (float n[3]) to return the cross product in
 */
static void
glmCross(float* u, float* v, float* n)
{
    //assert(u); assert(v); assert(n);
    
    n[0] = u[1]*v[2] - u[2]*v[1];
    n[1] = u[2]*v[0] - u[0]*v[2];
    n[2] = u[0]*v[1] - u[1]*v[0];
}

/* glmNormalize: normalize a vector
 *
 * v - array of 3 floats (float v[3]) to be normalized
 */
static void
glmNormalize(float* v)
{
    float l;
    
    //assert(v);
    
    l = (float)sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
    v[0] /= l;
    v[1] /= l;
    v[2] /= l;
}

/* glmEqual: compares two vectors and returns GL_TRUE if they are
 * equal (within a certain threshold) or GL_FALSE if not. An epsilon
 * that works fairly well is 0.000001.
 *
 * u - array of 3 floats (float u[3])
 * v - array of 3 floats (float v[3]) 
 */
static int
glmEqual(float* u, float* v, float epsilon)
{
    if (glmAbs(u[0] - v[0]) < epsilon &&
        glmAbs(u[1] - v[1]) < epsilon &&
        glmAbs(u[2] - v[2]) < epsilon) 
    {
        return TRUE;
    }
    return FALSE;
}

/* glmWeldVectors: eliminate (weld) vectors that are within an
 * epsilon of each other.
 *
 * vectors     - array of float[3]'s to be welded
 * numvectors - number of float[3]'s in vectors
 * epsilon     - maximum difference between vectors 
 *
 */
float*
glmWeldVectors(float* vectors, int* numvectors, float epsilon)
{
    float* copies;
    int   copied;
    int   i, j;
    
    copies = (float*)malloc(sizeof(float) * 3 * (*numvectors + 1));
    memcpy(copies, vectors, (sizeof(float) * 3 * (*numvectors + 1)));
    
    copied = 1;
    for (i = 1; i <= *numvectors; i++) {
        for (j = 1; j <= copied; j++) {
            if (glmEqual(&vectors[3 * i], &copies[3 * j], epsilon)) {
                goto duplicate;
            }
        }
        
        /* must not be any duplicates -- add to the copies array */
        copies[3 * copied + 0] = vectors[3 * i + 0];
        copies[3 * copied + 1] = vectors[3 * i + 1];
        copies[3 * copied + 2] = vectors[3 * i + 2];
        j = copied;             /* pass this along for below */
        copied++;
        
duplicate:
/* set the first component of this vector to point at the correct
        index into the new copies array */
        vectors[3 * i + 0] = (float)j;
    }
    
    *numvectors = copied-1;
    return copies;
}

/* glmFindGroup: Find a group in the model */
GLMgroup*
glmFindGroup(GLMmodel* model, char* name)
{
    GLMgroup* group;
    
    //assert(model);
    
    group = model->groups;
    while(group) {
        if (!strcmp(name, group->name))
            break;
        group = group->next;
    }
    
    return group;
}

/* glmAddGroup: Add a group to the model */
GLMgroup*
glmAddGroup(GLMmodel* model, char* name)
{
    GLMgroup* group;
    
    group = glmFindGroup(model, name);
    if (!group) {
        group = (GLMgroup*)malloc(sizeof(GLMgroup));
        group->name = _mystrdup(name);
        group->materials = NULL;
        group->numtriangles = 0;
        group->triangles = NULL;
        group->next = model->groups;
        model->groups = group;
        model->numgroups++;
    }
    return group;
}

/* glmFindGroup: Find a material in the model */
int
glmFindMaterial(GLMmodel* model, char* name)
{
    int i;
    
    /* XXX doing a linear search on a string key'd list is pretty lame,
    but it works and is fast enough for now. */
    for (i = 0; i < model->nummaterials; i++) {
        if (!strcmp(model->materials[i].name, name))
            goto found;
    }
    
    /* didn't find the name, so print a warning and return the default
    material (0). */
    //printf("glmFindMaterial():  can't find material \"%s\".\n", name);
    i = 0;
    
found:
    return i;
}


/* glmDirName: return the directory given a path
 *
 * path - filesystem path
 *
 * NOTE: the return value should be free'd.
 */
static char*
glmDirName(char* path)
{
    char* dir;
    char* s;
    
    dir = _mystrdup(path);
    
    s = strrchr(dir, '\\');
    if (s)
        s[1] = '\0';
    else
        dir[0] = '\0';
    
    return dir;
}


/* glmReadMTL: read a wavefront material library file
 *
 * model - properly initialized GLMmodel structure
 * name  - name of the material library
 */
static void
glmReadMTL(GLMmodel* model, char* name)
{
    FILE* file;
    char* dir;
    char* filename;
    char    buf[128];
    int nummaterials, i;
    
    dir = glmDirName(model->pathname);
    filename = (char*)malloc(sizeof(char) * (strlen(dir) + strlen(name) + 1));
    strcpy(filename, dir);
    strcat(filename, name);
    free(dir);
    
    file = fopen(filename, "r");
    if (!file) {
        //fprintf(stderr, "glmReadMTL() failed: can't open material file \"%s\".\n",
          //  filename);
        return;
        //exit(1);
    }
    free(filename);
    
    /* count the number of materials in the file */
    nummaterials = 1;
    while(fscanf(file, "%s", buf) != EOF) {
        switch(buf[0]) {
        case '#':               /* comment */
            /* eat up rest of line */
            fgets(buf, sizeof(buf), file);
            break;
        case 'n':               /* newmtl */
            fgets(buf, sizeof(buf), file);
            nummaterials++;
            sscanf(buf, "%s %s", buf, buf);
            break;
        default:
            /* eat up rest of line */
            fgets(buf, sizeof(buf), file);
            break;
        }
    }
    
    rewind(file);
    
    model->materials = (GLMmaterial*)malloc(sizeof(GLMmaterial) * nummaterials);
    model->nummaterials = nummaterials;
    
    /* set the default material */
    for (i = 0; i < nummaterials; i++) {
        model->materials[i].name = NULL;
        model->materials[i].shininess = 65.0f;
        model->materials[i].diffuse[0] = 0.8f;
        model->materials[i].diffuse[1] = 0.8f;
        model->materials[i].diffuse[2] = 0.8f;
        model->materials[i].diffuse[3] = 1.0f;
        model->materials[i].ambient[0] = 0.2f;
        model->materials[i].ambient[1] = 0.2f;
        model->materials[i].ambient[2] = 0.2f;
        model->materials[i].ambient[3] = 1.0f;
        model->materials[i].specular[0] = 0.0f;
        model->materials[i].specular[1] = 0.0f;
        model->materials[i].specular[2] = 0.0f;
        model->materials[i].specular[3] = 1.0f;
    }
    model->materials[0].name = _mystrdup("______default");
    
    /* now, read in the data */
    nummaterials = 0;
    while(fscanf(file, "%s", buf) != EOF) {
        switch(buf[0]) {
        case '#':               /* comment */
            /* eat up rest of line */
            fgets(buf, sizeof(buf), file);
            break;
        case 'n':               /* newmtl */
            fgets(buf, sizeof(buf), file);
            sscanf(buf, "%s %s", buf, buf);
            nummaterials++;
            model->materials[nummaterials].name = _mystrdup(buf);
            break;
        case 'N':
            fscanf(file, "%f", &model->materials[nummaterials].shininess);
            /* wavefront shininess is from [0, 1000], so scale for OpenGL */
            model->materials[nummaterials].shininess /= 1000.0f;
            model->materials[nummaterials].shininess *= 1280.0f;
            break;
        case 'K':
            switch(buf[1]) {
            case 'd':
                fscanf(file, "%f %f %f",
                    &model->materials[nummaterials].diffuse[0],
                    &model->materials[nummaterials].diffuse[1],
                    &model->materials[nummaterials].diffuse[2]);
                break;
            case 's':
                fscanf(file, "%f %f %f",
                    &model->materials[nummaterials].specular[0],
                    &model->materials[nummaterials].specular[1],
                    &model->materials[nummaterials].specular[2]);
                break;
            case 'a':
                fscanf(file, "%f %f %f",
                    &model->materials[nummaterials].ambient[0],
                    &model->materials[nummaterials].ambient[1],
                    &model->materials[nummaterials].ambient[2]);
                break;
            default:
                /* eat up rest of line */
                fgets(buf, sizeof(buf), file);
                break;
            }
            break;
            default:
                /* eat up rest of line */
                fgets(buf, sizeof(buf), file);
                break;
        }
    }
}


/* glmFirstPass: first pass at a Wavefront OBJ file that gets all the
 * statistics of the model (such as #vertices, #normals, etc)
 *
 * model - properly initialized GLMmodel structure
 * file  - (fopen'd) file descriptor 
 */

/*
   FIXED: a single group can contains several materials
 */

static void
glmFirstPass(GLMmodel* model, FILE* file) 
{
    int  numvertices;        /* number of vertices in model */
    int  numnormals;         /* number of normals in model */
    int  numtexcoords;       /* number of texcoords in model */
    int  numtriangles;       /* number of triangles in model */
    GLMgroup* group;            /* current group */
    unsigned    v, n, t;
    char        buf[128];
    int groupnamelen;

    /* make a default group */
    group = glmAddGroup(model, "______default");
    
    numvertices = numnormals = numtexcoords = numtriangles = 0;
    while(fscanf(file, "%s", buf) != EOF) {
        switch(buf[0]) {
        case '#':               /* comment */
            /* eat up rest of line */
            fgets(buf, sizeof(buf), file);
            break;
        case 'v':               /* v, vn, vt */
            switch(buf[1]) {
            case '\0':          /* vertex */
                /* eat up rest of line */
                fgets(buf, sizeof(buf), file);
                numvertices++;
                break;
            case 'n':           /* normal */
                /* eat up rest of line */
                fgets(buf, sizeof(buf), file);
                numnormals++;
                break;
            case 't':           /* texcoord */
                /* eat up rest of line */
                fgets(buf, sizeof(buf), file);
                numtexcoords++;
                break;
            default:
                //printf("glmFirstPass(): Unknown token \"%s\".\n", buf);
                //exit(1);
                return;
                //break;
            }
            break;
            case 'm':
                fgets(buf, sizeof(buf), file);
                sscanf(buf, "%s %s", buf, buf);
                model->mtllibname = _mystrdup(buf);
                glmReadMTL(model, buf);
                break;
            case 'u':
                /* eat up rest of line */
                fgets(buf, sizeof(buf), file);
                break;
            case 'g':               /* group */
                // Groups materials gathered at the end of the file

                /* eat up rest of line */
                fgets(buf, sizeof(buf), file);
                sscanf(buf, "%s", buf);
                groupnamelen = strlen(buf) - 1;
                if(buf[groupnamelen] == '\n') buf[groupnamelen] = '\0';  /* nuke '\n' */
                group = glmAddGroup(model, buf);
                break;
            case 'f':               /* face */
                v = n = t = 0;
                fscanf(file, "%s", buf);
                /* can be one of %d, %d//%d, %d/%d, %d/%d/%d %d//%d */
                if (strstr(buf, "//")) {
                    /* v//n */
                    sscanf(buf, "%d//%d", &v, &n);
                    fscanf(file, "%d//%d", &v, &n);
                    fscanf(file, "%d//%d", &v, &n);
                    numtriangles++;
                    group->numtriangles++;
                    while(fscanf(file, "%d//%d", &v, &n) > 0) {
                        numtriangles++;
                        group->numtriangles++;
                    }
                } else if (sscanf(buf, "%d/%d/%d", &v, &t, &n) == 3) {
                    /* v/t/n */
                    fscanf(file, "%d/%d/%d", &v, &t, &n);
                    fscanf(file, "%d/%d/%d", &v, &t, &n);
                    numtriangles++;
                    group->numtriangles++;
                    while(fscanf(file, "%d/%d/%d", &v, &t, &n) > 0) {
                        numtriangles++;
                        group->numtriangles++;
                    }
                } else if (sscanf(buf, "%d/%d", &v, &t) == 2) {
                    /* v/t */
                    fscanf(file, "%d/%d", &v, &t);
                    fscanf(file, "%d/%d", &v, &t);
                    numtriangles++;
                    group->numtriangles++;
                    while(fscanf(file, "%d/%d", &v, &t) > 0) {
                        numtriangles++;
                        group->numtriangles++;
                    }
                } else {
                    /* v */
                    fscanf(file, "%d", &v);
                    fscanf(file, "%d", &v);
                    numtriangles++;
                    group->numtriangles++;
                    while(fscanf(file, "%d", &v) > 0) {
                        numtriangles++;
                        group->numtriangles++;
                    }
                }
                break;
                
            default:
                /* eat up rest of line */
                fgets(buf, sizeof(buf), file);
                break;
        }
  }
  
  /* set the stats in the model structure */
  model->numvertices  = numvertices;
  model->numnormals   = numnormals;
  model->numtexcoords = numtexcoords;
  model->numtriangles = numtriangles;
  
  /* allocate memory for the triangles in each group */
  group = model->groups;
  while(group) {
      group->triangles = (int *) malloc(sizeof(int) * group->numtriangles);
      group->materials = (int *) malloc(sizeof(int) * group->numtriangles);
      group->numtriangles = 0;
      group = group->next;
  }
}

/* glmSecondPass: second pass at a Wavefront OBJ file that gets all
 * the data.
 *
 * model - properly initialized GLMmodel structure
 * file  - (fopen'd) file descriptor 
 */
static void
glmSecondPass(GLMmodel* model, FILE* file) 
{
    int  numvertices;        /* number of vertices in model */
    int  numnormals;         /* number of normals in model */
    int  numtexcoords;       /* number of texcoords in model */
    int  numtriangles;       /* number of triangles in model */
    float*    vertices;           /* array of vertices  */
    float*    normals;            /* array of normals */
    float*    texcoords;          /* array of texture coordinates */
    GLMgroup* group;            /* current group pointer */
    int  material;           /* current material */
    int  v, n, t;
    char        buf[128];
    int groupnamelen;

    /* set the pointer shortcuts */
    vertices       = model->vertices;
    normals    = model->normals;
    texcoords    = model->texcoords;
    group      = model->groups;
    
    /* on the second pass through the file, read all the data into the
    allocated arrays */
    numvertices = numnormals = numtexcoords = 1;
    numtriangles = 0;
    material = 0;
    while(fscanf(file, "%s", buf) != EOF) {
        switch(buf[0]) {
        case '#':               /* comment */
            /* eat up rest of line */
            fgets(buf, sizeof(buf), file);
            break;
        case 'v':               /* v, vn, vt */
            switch(buf[1]) {
            case '\0':          /* vertex */
                fscanf(file, "%f %f %f", 
                    &vertices[3 * numvertices + 0], 
                    &vertices[3 * numvertices + 1], 
                    &vertices[3 * numvertices + 2]);
                numvertices++;
                break;
            case 'n':           /* normal */
                fscanf(file, "%f %f %f", 
                    &normals[3 * numnormals + 0],
                    &normals[3 * numnormals + 1], 
                    &normals[3 * numnormals + 2]);
                numnormals++;
                break;
            case 't':           /* texcoord */
                fscanf(file, "%f %f", 
                    &texcoords[2 * numtexcoords + 0],
                    &texcoords[2 * numtexcoords + 1]);
                numtexcoords++;
                break;
            }
            break;
            case 'u':
                fgets(buf, sizeof(buf), file);
                sscanf(buf, "%s %s", buf, buf);
                material = glmFindMaterial(model, buf);
                break;
            case 'g':               /* group */

                // Groups materials gathered at the end of the file
                // (Apparently obsolete).
//                if(strcmp(buf, "group") == 0) {
  //                  groupnamelen = 1;
    //            }

                /* eat up rest of line */
                fgets(buf, sizeof(buf), file);
                sscanf(buf, "%s", buf);
                groupnamelen = strlen(buf) - 1;
                if(buf[groupnamelen] == '\n') buf[groupnamelen] = '\0';  /* nuke '\n' */
                group = glmFindGroup(model, buf);
                break;
            case 'f':               /* face */
                v = n = t = 0;
                fscanf(file, "%s", buf);
                /* can be one of %d, %d//%d, %d/%d, %d/%d/%d %d//%d */
                if (strstr(buf, "//")) {
                    /* v//n */
                    sscanf(buf, "%d//%d", &v, &n);
                    T(numtriangles).vindices[0] = v;
                    T(numtriangles).nindices[0] = n;
                    fscanf(file, "%d//%d", &v, &n);
                    T(numtriangles).vindices[1] = v;
                    T(numtriangles).nindices[1] = n;
                    fscanf(file, "%d//%d", &v, &n);
                    T(numtriangles).vindices[2] = v;
                    T(numtriangles).nindices[2] = n;
                    group->materials[group->numtriangles] = material;
                    group->triangles[group->numtriangles++] = numtriangles;
                    numtriangles++;
                    while(fscanf(file, "%d//%d", &v, &n) > 0) {
                        T(numtriangles).vindices[0] = T(numtriangles-1).vindices[0];
                        T(numtriangles).nindices[0] = T(numtriangles-1).nindices[0];
                        T(numtriangles).vindices[1] = T(numtriangles-1).vindices[2];
                        T(numtriangles).nindices[1] = T(numtriangles-1).nindices[2];
                        T(numtriangles).vindices[2] = v;
                        T(numtriangles).nindices[2] = n;
                        group->materials[group->numtriangles] = material;
                        group->triangles[group->numtriangles++] = numtriangles;
                        numtriangles++;
                    }
                
                } else if (sscanf(buf, "%d/%d/%d", &v, &t, &n) == 3) {
                    /* v/t/n */
                    T(numtriangles).vindices[0] = v;
                    T(numtriangles).tindices[0] = t;
                    T(numtriangles).nindices[0] = n;
                    fscanf(file, "%d/%d/%d", &v, &t, &n);
                    T(numtriangles).vindices[1] = v;
                    T(numtriangles).tindices[1] = t;
                    T(numtriangles).nindices[1] = n;
                    fscanf(file, "%d/%d/%d", &v, &t, &n);
                    T(numtriangles).vindices[2] = v;
                    T(numtriangles).tindices[2] = t;
                    T(numtriangles).nindices[2] = n;
                    group->materials[group->numtriangles] = material;
                    group->triangles[group->numtriangles++] = numtriangles;
                    numtriangles++;
                    while(fscanf(file, "%d/%d/%d", &v, &t, &n) > 0) {
                        T(numtriangles).vindices[0] = T(numtriangles-1).vindices[0];
                        T(numtriangles).tindices[0] = T(numtriangles-1).tindices[0];
                        T(numtriangles).nindices[0] = T(numtriangles-1).nindices[0];
                        T(numtriangles).vindices[1] = T(numtriangles-1).vindices[2];
                        T(numtriangles).tindices[1] = T(numtriangles-1).tindices[2];
                        T(numtriangles).nindices[1] = T(numtriangles-1).nindices[2];
                        T(numtriangles).vindices[2] = v;
                        T(numtriangles).tindices[2] = t;
                        T(numtriangles).nindices[2] = n;
                        group->materials[group->numtriangles] = material;
                        group->triangles[group->numtriangles++] = numtriangles;
                        numtriangles++;
                    }
                
                } else if (sscanf(buf, "%d/%d", &v, &t) == 2) {
                    /* v/t */
                    T(numtriangles).vindices[0] = v;
                    T(numtriangles).tindices[0] = t;
                    fscanf(file, "%d/%d", &v, &t);
                    T(numtriangles).vindices[1] = v;
                    T(numtriangles).tindices[1] = t;
                    fscanf(file, "%d/%d", &v, &t);
                    T(numtriangles).vindices[2] = v;
                    T(numtriangles).tindices[2] = t;
                    group->materials[group->numtriangles] = material;
                    group->triangles[group->numtriangles++] = numtriangles;
                    numtriangles++;
                    while(fscanf(file, "%d/%d", &v, &t) > 0) {
                        T(numtriangles).vindices[0] = T(numtriangles-1).vindices[0];
                        T(numtriangles).tindices[0] = T(numtriangles-1).tindices[0];
                        T(numtriangles).vindices[1] = T(numtriangles-1).vindices[2];
                        T(numtriangles).tindices[1] = T(numtriangles-1).tindices[2];
                        T(numtriangles).vindices[2] = v;
                        T(numtriangles).tindices[2] = t;
                        group->materials[group->numtriangles] = material;
                        group->triangles[group->numtriangles++] = numtriangles;
                        numtriangles++;
                    }
                
                } else {
                    /* v */
                    sscanf(buf, "%d", &v);
                    T(numtriangles).vindices[0] = v;
                    fscanf(file, "%d", &v);
                    T(numtriangles).vindices[1] = v;
                    fscanf(file, "%d", &v);
                    T(numtriangles).vindices[2] = v;
                    group->materials[group->numtriangles] = material;
                    group->triangles[group->numtriangles++] = numtriangles;
                    numtriangles++;
                    while(fscanf(file, "%d", &v) > 0) {
                        T(numtriangles).vindices[0] = T(numtriangles-1).vindices[0];
                        T(numtriangles).vindices[1] = T(numtriangles-1).vindices[2];
                        T(numtriangles).vindices[2] = v;
                        group->materials[group->numtriangles] = material;
                        group->triangles[group->numtriangles++] = numtriangles;
                        numtriangles++;
                    }
                    
                }
                break;
                
            default:
                /* eat up rest of line */
                fgets(buf, sizeof(buf), file);
                break;
    }
  }
  
//#if 0
//  /* announce the memory requirements */
//  printf(" Memory: %d bytes\n",
      //numvertices  * 3*sizeof(float) +
      //numnormals   * 3*sizeof(float) * (numnormals ? 1 : 0) +
      //numtexcoords * 3*sizeof(float) * (numtexcoords ? 1 : 0) +
      //numtriangles * sizeof(GLMtriangle));
//#endif
}


/* public functions */


/* glmUnitize: "unitize" a model by translating it to the origin and
 * scaling it to fit in a unit cube around the origin.   Returns the
 * scalefactor used.
 *
 * model - properly initialized GLMmodel structure 
 */
float __stdcall
glmUnitize(GLMmodel* model)
{
    int  i;
    float maxx, minx, maxy, miny, maxz, minz;
    float cx, cy, cz, w, h, d;
    float scale;
    
    //assert(model);
    //assert(model->vertices);
    
    /* get the max/mins */
    maxx = minx = model->vertices[3 + 0];
    maxy = miny = model->vertices[3 + 1];
    maxz = minz = model->vertices[3 + 2];
    for (i = 1; i <= model->numvertices; i++) {
        if (maxx < model->vertices[3 * i + 0])
            maxx = model->vertices[3 * i + 0];
        if (minx > model->vertices[3 * i + 0])
            minx = model->vertices[3 * i + 0];
        
        if (maxy < model->vertices[3 * i + 1])
            maxy = model->vertices[3 * i + 1];
        if (miny > model->vertices[3 * i + 1])
            miny = model->vertices[3 * i + 1];
        
        if (maxz < model->vertices[3 * i + 2])
            maxz = model->vertices[3 * i + 2];
        if (minz > model->vertices[3 * i + 2])
            minz = model->vertices[3 * i + 2];
    }
    
    /* calculate model width, height, and depth */
    w = glmAbs(maxx) + glmAbs(minx);
    h = glmAbs(maxy) + glmAbs(miny);
    d = glmAbs(maxz) + glmAbs(minz);
    
    /* calculate center of the model */
    cx = (maxx + minx) / 2.0f;
    cy = (maxy + miny) / 2.0f;
    cz = (maxz + minz) / 2.0f;
    
    /* calculate unitizing scale factor */
    scale = 2.0f / glmMax(glmMax(w, h), d);
    
    /* translate around center then scale */
    for (i = 1; i <= model->numvertices; i++) {
        model->vertices[3 * i + 0] -= cx;
        model->vertices[3 * i + 1] -= cy;
        model->vertices[3 * i + 2] -= cz;
        model->vertices[3 * i + 0] *= scale;
        model->vertices[3 * i + 1] *= scale;
        model->vertices[3 * i + 2] *= scale;
    }
    
    return scale;
}

/* glmDimensions: Calculates the dimensions (width, height, depth) of
 * a model.
 *
 * model   - initialized GLMmodel structure
 * dimensions - array of 3 floats (float dimensions[3])
 */
void __stdcall
glmDimensions(GLMmodel* model, float* dimensions)
{
    int i;
    float maxx, minx, maxy, miny, maxz, minz;
    
    //assert(model);
    //assert(model->vertices);
    //assert(dimensions);
    
    /* get the max/mins */
    maxx = minx = model->vertices[3 + 0];
    maxy = miny = model->vertices[3 + 1];
    maxz = minz = model->vertices[3 + 2];
    for (i = 1; i <= model->numvertices; i++) {
        if (maxx < model->vertices[3 * i + 0])
            maxx = model->vertices[3 * i + 0];
        if (minx > model->vertices[3 * i + 0])
            minx = model->vertices[3 * i + 0];
        
        if (maxy < model->vertices[3 * i + 1])
            maxy = model->vertices[3 * i + 1];
        if (miny > model->vertices[3 * i + 1])
            miny = model->vertices[3 * i + 1];
        
        if (maxz < model->vertices[3 * i + 2])
            maxz = model->vertices[3 * i + 2];
        if (minz > model->vertices[3 * i + 2])
            minz = model->vertices[3 * i + 2];
    }
    
    /* calculate model width, height, and depth */
    dimensions[0] = glmAbs(maxx) + glmAbs(minx);
    dimensions[1] = glmAbs(maxy) + glmAbs(miny);
    dimensions[2] = glmAbs(maxz) + glmAbs(minz);
}

/* glmScale: Scales a model by a given amount.
 * 
 * model - properly initialized GLMmodel structure
 * scale - scalefactor (0.5 = half as large, 2.0 = twice as large)
 */
void __stdcall
glmScale(GLMmodel* model, float scale)
{
    int i;
    
    for (i = 1; i <= model->numvertices; i++) {
        model->vertices[3 * i + 0] *= scale;
        model->vertices[3 * i + 1] *= scale;
        model->vertices[3 * i + 2] *= scale;
    }
}

/* glmReverseWinding: Reverse the polygon winding for all polygons in
 * this model.   Default winding is counter-clockwise.  Also changes
 * the direction of the normals.
 * 
 * model - properly initialized GLMmodel structure 
 */
void __stdcall
glmReverseWinding(GLMmodel* model)
{
    int i, swap;
    
    //assert(model);
    
    for (i = 0; i < model->numtriangles; i++) {
        swap = T(i).vindices[0];
        T(i).vindices[0] = T(i).vindices[2];
        T(i).vindices[2] = swap;
        
        if (model->numnormals) {
            swap = T(i).nindices[0];
            T(i).nindices[0] = T(i).nindices[2];
            T(i).nindices[2] = swap;
        }
        
        if (model->numtexcoords) {
            swap = T(i).tindices[0];
            T(i).tindices[0] = T(i).tindices[2];
            T(i).tindices[2] = swap;
        }
    }
    
    /* reverse facet normals */
    for (i = 1; i <= model->numfacetnorms; i++) {
        model->facetnorms[3 * i + 0] = -model->facetnorms[3 * i + 0];
        model->facetnorms[3 * i + 1] = -model->facetnorms[3 * i + 1];
        model->facetnorms[3 * i + 2] = -model->facetnorms[3 * i + 2];
    }
    
    /* reverse vertex normals */
    for (i = 1; i <= model->numnormals; i++) {
        model->normals[3 * i + 0] = -model->normals[3 * i + 0];
        model->normals[3 * i + 1] = -model->normals[3 * i + 1];
        model->normals[3 * i + 2] = -model->normals[3 * i + 2];
    }
}

/* glmFacetNormals: Generates facet normals for a model (by taking the
 * cross product of the two vectors derived from the sides of each
 * triangle).  Assumes a counter-clockwise winding.
 *
 * model - initialized GLMmodel structure
 */
void __stdcall
glmFacetNormals(GLMmodel* model)
{
    int  i;
    float u[3];
    float v[3];
    
    //assert(model);
    //assert(model->vertices);
    
    /* clobber any old facetnormals */
    if (model->facetnorms)
        free(model->facetnorms);
    
    /* allocate memory for the new facet normals */
    model->numfacetnorms = model->numtriangles;
    model->facetnorms = (float*)malloc(sizeof(float) *
                       3 * (model->numfacetnorms + 1));
    
    for (i = 0; i < model->numtriangles; i++) {
        model->triangles[i].findex = i+1;
        
        u[0] = model->vertices[3 * T(i).vindices[1] + 0] -
            model->vertices[3 * T(i).vindices[0] + 0];
        u[1] = model->vertices[3 * T(i).vindices[1] + 1] -
            model->vertices[3 * T(i).vindices[0] + 1];
        u[2] = model->vertices[3 * T(i).vindices[1] + 2] -
            model->vertices[3 * T(i).vindices[0] + 2];
        
        v[0] = model->vertices[3 * T(i).vindices[2] + 0] -
            model->vertices[3 * T(i).vindices[0] + 0];
        v[1] = model->vertices[3 * T(i).vindices[2] + 1] -
            model->vertices[3 * T(i).vindices[0] + 1];
        v[2] = model->vertices[3 * T(i).vindices[2] + 2] -
            model->vertices[3 * T(i).vindices[0] + 2];
        
        glmCross(u, v, &model->facetnorms[3 * (i+1)]);
        glmNormalize(&model->facetnorms[3 * (i+1)]);
    }
}

/* glmVertexNormals: Generates smooth vertex normals for a model.
 * First builds a list of all the triangles each vertex is in.   Then
 * loops through each vertex in the the list averaging all the facet
 * normals of the triangles each vertex is in.   Finally, sets the
 * normal index in the triangle for the vertex to the generated smooth
 * normal.   If the dot product of a facet normal and the facet normal
 * associated with the first triangle in the list of triangles the
 * current vertex is in is greater than the cosine of the angle
 * parameter to the function, that facet normal is not added into the
 * average normal calculation and the corresponding vertex is given
 * the facet normal.  This tends to preserve hard edges.  The angle to
 * use depends on the model, but 90 degrees is usually a good start.
 *
 * model - initialized GLMmodel structure
 * angle - maximum angle (in degrees) to smooth across
 */
void __stdcall
glmVertexNormals(GLMmodel* model, float angle)
{
    GLMnode*    node;
    GLMnode*    tail;
    GLMnode** members;
    float*    normals;
    int  numnormals;
    float average[3];
    float dot, cos_angle;
    int  i, avg;
    
    //assert(model);
    //assert(model->facetnorms);
    
    /* calculate the cosine of the angle (in degrees) */
    cos_angle = (float) cos(angle * M_PI / 180.0f);
    
    /* nuke any previous normals */
    if (model->normals)
        free(model->normals);
    
    /* allocate space for new normals */
    model->numnormals = model->numtriangles * 3; /* 3 normals per triangle */
    model->normals = (float*)malloc(sizeof(float)* 3* (model->numnormals+1));
    
    /* allocate a structure that will hold a linked list of triangle
    indices for each vertex */
    members = (GLMnode**)malloc(sizeof(GLMnode*) * (model->numvertices + 1));
    for (i = 1; i <= model->numvertices; i++)
        members[i] = NULL;
    
    /* for every triangle, create a node for each vertex in it */
    for (i = 0; i < model->numtriangles; i++) {
        node = (GLMnode*)malloc(sizeof(GLMnode));
        node->index = i;
        node->next  = members[T(i).vindices[0]];
        members[T(i).vindices[0]] = node;
        
        node = (GLMnode*)malloc(sizeof(GLMnode));
        node->index = i;
        node->next  = members[T(i).vindices[1]];
        members[T(i).vindices[1]] = node;
        
        node = (GLMnode*)malloc(sizeof(GLMnode));
        node->index = i;
        node->next  = members[T(i).vindices[2]];
        members[T(i).vindices[2]] = node;
    }
    
    /* calculate the average normal for each vertex */
    numnormals = 1;
    for (i = 1; i <= model->numvertices; i++) {
    /* calculate an average normal for this vertex by averaging the
        facet normal of every triangle this vertex is in */
        node = members[i];
        //if (!node)
        //    fprintf(stderr, "glmVertexNormals(): vertex w/o a triangle\n");
        average[0] = 0.0; average[1] = 0.0; average[2] = 0.0;
        avg = 0;
        while (node) {
        /* only average if the dot product of the angle between the two
        facet normals is greater than the cosine of the threshold
        angle -- or, said another way, the angle between the two
            facet normals is less than (or equal to) the threshold angle */
            dot = glmDot(&model->facetnorms[3 * T(node->index).findex],
                &model->facetnorms[3 * T(members[i]->index).findex]);
            if (dot > cos_angle) {
                node->averaged = TRUE;
                average[0] += model->facetnorms[3 * T(node->index).findex + 0];
                average[1] += model->facetnorms[3 * T(node->index).findex + 1];
                average[2] += model->facetnorms[3 * T(node->index).findex + 2];
                avg = 1;            /* we averaged at least one normal! */
            } else {
                node->averaged = FALSE;
            }
            node = node->next;
        }
        
        if (avg) {
            /* normalize the averaged normal */
            glmNormalize(average);
            
            /* add the normal to the vertex normals list */
            model->normals[3 * numnormals + 0] = average[0];
            model->normals[3 * numnormals + 1] = average[1];
            model->normals[3 * numnormals + 2] = average[2];
            avg = numnormals;
            numnormals++;
        }
        
        /* set the normal of this vertex in each triangle it is in */
        node = members[i];
        while (node) {
            if (node->averaged) {
                /* if this node was averaged, use the average normal */
                if (T(node->index).vindices[0] == i)
                    T(node->index).nindices[0] = avg;
                else if (T(node->index).vindices[1] == i)
                    T(node->index).nindices[1] = avg;
                else if (T(node->index).vindices[2] == i)
                    T(node->index).nindices[2] = avg;
            } else {
                /* if this node wasn't averaged, use the facet normal */
                model->normals[3 * numnormals + 0] = 
                    model->facetnorms[3 * T(node->index).findex + 0];
                model->normals[3 * numnormals + 1] = 
                    model->facetnorms[3 * T(node->index).findex + 1];
                model->normals[3 * numnormals + 2] = 
                    model->facetnorms[3 * T(node->index).findex + 2];
                if (T(node->index).vindices[0] == i)
                    T(node->index).nindices[0] = numnormals;
                else if (T(node->index).vindices[1] == i)
                    T(node->index).nindices[1] = numnormals;
                else if (T(node->index).vindices[2] == i)
                    T(node->index).nindices[2] = numnormals;
                numnormals++;
            }
            node = node->next;
        }
    }
    
    model->numnormals = numnormals - 1;
    
    /* free the member information */
    for (i = 1; i <= model->numvertices; i++) {
        node = members[i];
        while (node) {
            tail = node;
            node = node->next;
            free(tail);
        }
    }
    free(members);
    
    /* pack the normals array (we previously allocated the maximum
    number of normals that could possibly be created (numtriangles *
    3), so get rid of some of them (usually alot unless none of the
    facet normals were averaged)) */
    normals = model->normals;
    model->normals = (float*)malloc(sizeof(float)* 3* (model->numnormals+1));
    for (i = 1; i <= model->numnormals; i++) {
        model->normals[3 * i + 0] = normals[3 * i + 0];
        model->normals[3 * i + 1] = normals[3 * i + 1];
        model->normals[3 * i + 2] = normals[3 * i + 2];
    }
    free(normals);
}


/* glmLinearTexture: Generates texture coordinates according to a
 * linear projection of the texture map.  It generates these by
 * linearly mapping the vertices onto a square.
 *
 * model - pointer to initialized GLMmodel structure
 */
void __stdcall
glmLinearTexture(GLMmodel* model)
{
    GLMgroup *group;
    float dimensions[3];
    float x, y, scalefactor;
    int i;
    
    //assert(model);
    
    if (model->texcoords)
        free(model->texcoords);
    model->numtexcoords = model->numvertices;
    model->texcoords=(float*)malloc(sizeof(float)*2*(model->numtexcoords+1));
    
    glmDimensions(model, dimensions);
    scalefactor = 2.0f / glmAbs(glmMax(glmMax(dimensions[0], dimensions[1]), dimensions[2]));
    
    /* do the calculations */
    for(i = 1; i <= model->numvertices; i++) {
        x = model->vertices[3 * i + 0] * scalefactor;
        y = model->vertices[3 * i + 2] * scalefactor;
        model->texcoords[2 * i + 0] = (x + 1.0f) / 2.0f;
        model->texcoords[2 * i + 1] = (y + 1.0f) / 2.0f;
    }
    
    /* go through and put texture coordinate indices in all the triangles */
    group = model->groups;
    while(group) {
        for(i = 0; i < group->numtriangles; i++) {
            T(group->triangles[i]).tindices[0] = T(group->triangles[i]).vindices[0];
            T(group->triangles[i]).tindices[1] = T(group->triangles[i]).vindices[1];
            T(group->triangles[i]).tindices[2] = T(group->triangles[i]).vindices[2];
        }    
        group = group->next;
    }
    
//#if 0
//    printf("glmLinearTexture(): generated %d linear texture coordinates\n",
//        model->numtexcoords);
//#endif
}

/* glmSpheremapTexture: Generates texture coordinates according to a
 * spherical projection of the texture map.  Sometimes referred to as
 * spheremap, or reflection map texture coordinates.  It generates
 * these by using the normal to calculate where that vertex would map
 * onto a sphere.  Since it is impossible to map something flat
 * perfectly onto something spherical, there is distortion at the
 * poles.  This particular implementation causes the poles along the X
 * axis to be distorted.
 *
 * model - pointer to initialized GLMmodel structure
 */
void __stdcall
glmSpheremapTexture(GLMmodel* model)
{
    GLMgroup* group;
    float theta, phi, rho, x, y, z, r;
    int i;
    
    //assert(model);
    //assert(model->normals);
    
    if (model->texcoords) free(model->texcoords);
    model->numtexcoords = model->numnormals;
    model->texcoords=(float*)malloc(sizeof(float)*2*(model->numtexcoords+1));
    
    for (i = 1; i <= model->numnormals; i++) {
        z = model->normals[3 * i + 0];  /* re-arrange for pole distortion */
        y = model->normals[3 * i + 1];
        x = model->normals[3 * i + 2];
        r = (float) sqrt((x * x) + (y * y));
        rho = (float) sqrt((r * r) + (z * z));
        
        if(r == 0.0f) {
            theta = 0.0f;
            phi = 0.0f;
        } else {
            if(z == 0.0f)
                phi = 3.14159265f / 2.0f;
            else
                phi = (float) acos(z / rho);
            
            if(y == 0.0f)
                theta = 3.141592365f / 2.0f;
            else
                theta = (float) asin(y / r) + (3.14159265f / 2.0f);
        }
        
        model->texcoords[2 * i + 0] = theta / 3.14159265f;
        model->texcoords[2 * i + 1] = phi / 3.14159265f;
    }
    
    /* go through and put texcoord indices in all the triangles */
    group = model->groups;
    while(group) {
        for (i = 0; i < group->numtriangles; i++) {
            T(group->triangles[i]).tindices[0] = T(group->triangles[i]).nindices[0];
            T(group->triangles[i]).tindices[1] = T(group->triangles[i]).nindices[1];
            T(group->triangles[i]).tindices[2] = T(group->triangles[i]).nindices[2];
        }
        group = group->next;
    }
}

/* glmDelete: Deletes a GLMmodel structure.
 *
 * model - initialized GLMmodel structure
 */
void __stdcall
glmDelete(GLMmodel* model)
{
    GLMgroup* group;
    int i;
    
    if(model) {
        //assert(model);
    
        if (model->pathname) free(model->pathname);
        if (model->mtllibname) free(model->mtllibname);
        if (model->vertices) free(model->vertices);
        if (model->normals) free(model->normals);
        if (model->texcoords) free(model->texcoords);
        if (model->facetnorms) free(model->facetnorms);
        if (model->triangles) free(model->triangles);
        if (model->materials) {
            for (i = 0; i < model->nummaterials; i++)
                if(model->materials[i].name) free(model->materials[i].name);
        }
        if(model->materials) free(model->materials);
        while(model->groups) {
            group = model->groups;
            model->groups = model->groups->next;
            if(group->name) free(group->name);
            if(group->materials) free(group->materials);
            if(group->triangles) free(group->triangles);
            if(group) free(group);
        }
    
        free(model);
    }
}

/* glmReadOBJ: Reads a model description from a Wavefront .OBJ file.
 * Returns a pointer to the created object which should be free'd with
 * glmDelete().
 *
 * filename - name of the file containing the Wavefront .OBJ format data.  
 */
GLMmodel* __stdcall
glmReadOBJ(char* filename)
{
    GLMmodel* model;
    FILE*   file;
    
    /* open the file */
    file = fopen(filename, "r");
    if (!file) {
        //fprintf(stderr, "glmReadOBJ() failed: can't open data file \"%s\".\n",
          //  filename);
        return(NULL);
        //exit(1);
    }
    
    /* allocate a new model */
    model = (GLMmodel*)malloc(sizeof(GLMmodel));
    model->pathname    = _mystrdup(filename);
    model->mtllibname    = NULL;
    model->numvertices   = 0;
    model->vertices    = NULL;
    model->numnormals    = 0;
    model->normals     = NULL;
    model->numtexcoords  = 0;
    model->texcoords       = NULL;
    model->numfacetnorms = 0;
    model->facetnorms    = NULL;
    model->numtriangles  = 0;
    model->triangles       = NULL;
    model->nummaterials  = 0;
    model->materials       = NULL;
    model->numgroups       = 0;
    model->groups      = NULL;
    model->position[0]   = 0.0;
    model->position[1]   = 0.0;
    model->position[2]   = 0.0;
    
    /* make a first pass through the file to get a count of the number
    of vertices, normals, texcoords & triangles */
    glmFirstPass(model, file);
    
    /* allocate memory */
    model->vertices = (float*)malloc(sizeof(float) *
        3 * (model->numvertices + 1));
    model->triangles = (GLMtriangle*)malloc(sizeof(GLMtriangle) *
        model->numtriangles);
    if (model->numnormals) {
        model->normals = (float*)malloc(sizeof(float) *
            3 * (model->numnormals + 1));
    }
    if (model->numtexcoords) {
        model->texcoords = (float*)malloc(sizeof(float) *
            2 * (model->numtexcoords + 1));
    }
    
    /* rewind to beginning of file and read in the data this pass */
    rewind(file);
    
    glmSecondPass(model, file);
    
    /* close the file */
    fclose(file);
    
    return model;
}

/* glmDraw: Renders the model to the current OpenGL context using the
 * mode specified.
 *
 * model - initialized GLMmodel structure
 * mode  - a bitwise OR of values describing what is to be rendered.
 *             GLM_NONE     -  render with only vertices
 *             GLM_FLAT     -  render with facet normals
 *             GLM_SMOOTH   -  render with vertex normals
 *             GLM_TEXTURE  -  render with texture coords
 *             GLM_COLOR    -  render with colors (color material)
 *             GLM_MATERIAL -  render with materials
 *             GLM_COLOR and GLM_MATERIAL should not both be specified.  
 *             GLM_FLAT and GLM_SMOOTH should not both be specified.  
 */
void __stdcall
glmDraw(GLMmodel* model, int mode) {
    static int i;
    static GLMgroup* group;
    static GLMtriangle* triangle;
    static GLMmaterial* material;
    
    //assert(model);
    //assert(model->vertices);
    
    /* do a bit of warning */
    if (mode & GLM_FLAT && !model->facetnorms) {
        //printf("glmDraw() warning: flat render mode requested "
          //  "with no facet normals defined.\n");
        mode &= ~GLM_FLAT;
    }
    if (mode & GLM_SMOOTH && !model->normals) {
        //printf("glmDraw() warning: smooth render mode requested "
          //  "with no normals defined.\n");
        mode &= ~GLM_SMOOTH;
    }
    if (mode & GLM_TEXTURE && !model->texcoords) {
        //printf("glmDraw() warning: texture render mode requested "
          //  "with no texture coordinates defined.\n");
        mode &= ~GLM_TEXTURE;
    }
    if (mode & GLM_FLAT && mode & GLM_SMOOTH) {
        //printf("glmDraw() warning: flat render mode requested "
          //  "and smooth render mode requested (using smooth).\n");
        mode &= ~GLM_FLAT;
    }
    if (mode & GLM_COLOR && !model->materials) {
        //printf("glmDraw() warning: color render mode requested "
          //  "with no materials defined.\n");
        mode &= ~GLM_COLOR;
    }
    if (mode & GLM_MATERIAL && !model->materials) {
        //printf("glmDraw() warning: material render mode requested "
          //  "with no materials defined.\n");
        mode &= ~GLM_MATERIAL;
    }
    if (mode & GLM_COLOR && mode & GLM_MATERIAL) {
        //printf("glmDraw() warning: color and material render mode requested "
          //  "using only material mode.\n");
        mode &= ~GLM_COLOR;
    }
    if (mode & GLM_COLOR)
        glEnable(GL_COLOR_MATERIAL);
    else if (mode & GLM_MATERIAL)
        glDisable(GL_COLOR_MATERIAL);
    
    /* perhaps this loop should be unrolled into material, color, flat,
       smooth, etc. loops?  since most cpu's have good branch prediction
       schemes (and these branches will always go one way), probably
       wouldn't gain too much?  */
    
    group = model->groups;
    while (group) {
        for (i = 0; i < group->numtriangles; i++) {
            material = &model->materials[group->materials[i]];
            if (mode & GLM_MATERIAL) {
                glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, material->ambient);
                glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, material->diffuse);
                glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, material->specular);
                glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, material->shininess);
            }
        
            if (mode & GLM_COLOR) {
                glColor3fv(material->diffuse);
            }
        
            glBegin(GL_TRIANGLES);
                triangle = &T(group->triangles[i]);
            
                if (mode & GLM_FLAT) glNormal3fv(&model->facetnorms[3 * triangle->findex]);
            
                if (mode & GLM_SMOOTH) glNormal3fv(&model->normals[3 * triangle->nindices[0]]);
                if (mode & GLM_TEXTURE) glTexCoord2fv(&model->texcoords[2 * triangle->tindices[0]]);
                glVertex3fv(&model->vertices[3 * triangle->vindices[0]]);
            
                if (mode & GLM_SMOOTH) glNormal3fv(&model->normals[3 * triangle->nindices[1]]);
                if (mode & GLM_TEXTURE) glTexCoord2fv(&model->texcoords[2 * triangle->tindices[1]]);
                glVertex3fv(&model->vertices[3 * triangle->vindices[1]]);
            
                if (mode & GLM_SMOOTH) glNormal3fv(&model->normals[3 * triangle->nindices[2]]);
                if (mode & GLM_TEXTURE) glTexCoord2fv(&model->texcoords[2 * triangle->tindices[2]]);
                glVertex3fv(&model->vertices[3 * triangle->vindices[2]]);
            glEnd();
            
        }
        
        group = group->next;
    }
}

GLvoid __stdcall
glmCreateVertexArray(GLMmodel *model, int mode, GLMArray *VertexArray)
{
    static int i;
    static GLMgroup* group;
    static GLMtriangle* triangle;
    static GLMmaterial* material;
    int Max_Nbr_Triangles;
    GLMArray TmpVertexArray;
    GLMArray *TmpVertexArrayPtr;
    
    //assert(model);
    //assert(model->vertices);
    
    /* do a bit of warning */
    if (mode & GLM_FLAT && !model->facetnorms) {
        mode &= ~GLM_FLAT;
    }
    if (mode & GLM_SMOOTH && !model->normals) {
        mode &= ~GLM_SMOOTH;
    }
    if (mode & GLM_TEXTURE && !model->texcoords) {
        mode &= ~GLM_TEXTURE;
    }
    if (mode & GLM_FLAT && mode & GLM_SMOOTH) {
        mode &= ~GLM_FLAT;
    }
    if (mode & GLM_COLOR && !model->materials) {
        mode &= ~GLM_COLOR;
    }
    if (mode & GLM_MATERIAL && !model->materials) {
        mode &= ~GLM_MATERIAL;
    }
    if (mode & GLM_COLOR && mode & GLM_MATERIAL) {
        mode &= ~GLM_COLOR;
    }
    //if (mode & GLM_COLOR)
      //  glEnable(GL_COLOR_MATERIAL);
    //else if (mode & GLM_MATERIAL)
      //  glDisable(GL_COLOR_MATERIAL);
    
    group = model->groups;
    Max_Nbr_Triangles = 0;
    while (group) {
        Max_Nbr_Triangles += group->numtriangles;
        group = group->next;
    }
    VertexArray->Nbr_Triangles = Max_Nbr_Triangles * 3;

    VertexArray->Triangles = (float *) malloc(Max_Nbr_Triangles * sizeof(float) * 3 * 3);
    VertexArray->Normals = (float *) malloc(Max_Nbr_Triangles * sizeof(float)  * 3 * 3);
    VertexArray->Colors = (float *) malloc(Max_Nbr_Triangles * sizeof(float)  * 3 * 3);
    VertexArray->Textures = (float *) malloc(Max_Nbr_Triangles * sizeof(float)  * 2 * 3);
    TmpVertexArray.Normals = VertexArray->Normals;
    TmpVertexArray.Textures = VertexArray->Textures;
    TmpVertexArray.Triangles = VertexArray->Triangles;
    TmpVertexArray.Colors = VertexArray->Colors;
    TmpVertexArrayPtr = &TmpVertexArray;

    group = model->groups;
    while (group) {
        for (i = 0; i < group->numtriangles; i++) {
            material = &model->materials[group->materials[i]];
            *TmpVertexArrayPtr->Colors++ = material->diffuse[0];
            *TmpVertexArrayPtr->Colors++ = material->diffuse[1];
            *TmpVertexArrayPtr->Colors++ = material->diffuse[2];
        
            triangle = &T(group->triangles[i]);

            if (mode & GLM_FLAT) {
                *TmpVertexArrayPtr->Normals++ = model->facetnorms[(3 * triangle->findex)];
                *TmpVertexArrayPtr->Normals++ = model->facetnorms[(3 * triangle->findex) + 1];
                *TmpVertexArrayPtr->Normals++ = model->facetnorms[(3 * triangle->findex) + 2];
            } else {
                *TmpVertexArrayPtr->Normals++ = model->normals[(3 * triangle->nindices[0])];
                *TmpVertexArrayPtr->Normals++ = model->normals[(3 * triangle->nindices[0]) + 1];
                *TmpVertexArrayPtr->Normals++ = model->normals[(3 * triangle->nindices[0]) + 2];
            }        
            if(mode & GLM_TEXTURE) {
                *TmpVertexArrayPtr->Textures++ = model->texcoords[(2 * triangle->tindices[0])];
                *TmpVertexArrayPtr->Textures++ = model->texcoords[(2 * triangle->tindices[0]) + 1];
            } else {
                *TmpVertexArrayPtr->Textures++ = 0.0f;
                *TmpVertexArrayPtr->Textures++ = 0.0f;
            }

            *TmpVertexArrayPtr->Colors++ = material->diffuse[0];
            *TmpVertexArrayPtr->Colors++ = material->diffuse[1];
            *TmpVertexArrayPtr->Colors++ = material->diffuse[2];

            *TmpVertexArrayPtr->Triangles++ = model->vertices[(3 * triangle->vindices[0])];
            *TmpVertexArrayPtr->Triangles++ = model->vertices[(3 * triangle->vindices[0]) + 1];
            *TmpVertexArrayPtr->Triangles++ = model->vertices[(3 * triangle->vindices[0]) + 2];

            if (mode & GLM_FLAT) {
                *TmpVertexArrayPtr->Normals++ = model->facetnorms[(3 * triangle->findex)];
                *TmpVertexArrayPtr->Normals++ = model->facetnorms[(3 * triangle->findex) + 1];
                *TmpVertexArrayPtr->Normals++ = model->facetnorms[(3 * triangle->findex) + 2];
            } else {
                *TmpVertexArrayPtr->Normals++ = model->normals[(3 * triangle->nindices[1])];
                *TmpVertexArrayPtr->Normals++ = model->normals[(3 * triangle->nindices[1]) + 1];
                *TmpVertexArrayPtr->Normals++ = model->normals[(3 * triangle->nindices[1]) + 2];
            }
            if(mode & GLM_TEXTURE) {
                *TmpVertexArrayPtr->Textures++ = model->texcoords[(2 * triangle->tindices[1])];
                *TmpVertexArrayPtr->Textures++ = model->texcoords[(2 * triangle->tindices[1]) + 1];
            } else {
                *TmpVertexArrayPtr->Textures++ = 0.0f;
                *TmpVertexArrayPtr->Textures++ = 0.0f;
            }

            *TmpVertexArrayPtr->Colors++ = material->diffuse[0];
            *TmpVertexArrayPtr->Colors++ = material->diffuse[1];
            *TmpVertexArrayPtr->Colors++ = material->diffuse[2];

            *TmpVertexArrayPtr->Triangles++ = model->vertices[(3 * triangle->vindices[1])];
            *TmpVertexArrayPtr->Triangles++ = model->vertices[(3 * triangle->vindices[1]) + 1];
            *TmpVertexArrayPtr->Triangles++ = model->vertices[(3 * triangle->vindices[1]) + 2];

            if (mode & GLM_FLAT) {
                *TmpVertexArrayPtr->Normals++ = model->facetnorms[(3 * triangle->findex)];
                *TmpVertexArrayPtr->Normals++ = model->facetnorms[(3 * triangle->findex) + 1];
                *TmpVertexArrayPtr->Normals++ = model->facetnorms[(3 * triangle->findex) + 2];
            } else {
                *TmpVertexArrayPtr->Normals++ = model->normals[(3 * triangle->nindices[2])];
                *TmpVertexArrayPtr->Normals++ = model->normals[(3 * triangle->nindices[2]) + 1];
                *TmpVertexArrayPtr->Normals++ = model->normals[(3 * triangle->nindices[2]) + 2];
            }
            if(mode & GLM_TEXTURE) {
                *TmpVertexArrayPtr->Textures++ = model->texcoords[(2 * triangle->tindices[2])];
                *TmpVertexArrayPtr->Textures++ = model->texcoords[(2 * triangle->tindices[2]) + 1];
            } else {
                *TmpVertexArrayPtr->Textures++ = 0.0f;
                *TmpVertexArrayPtr->Textures++ = 0.0f;
            }

            *TmpVertexArrayPtr->Triangles++ = model->vertices[(3 * triangle->vindices[2])];
            *TmpVertexArrayPtr->Triangles++ = model->vertices[(3 * triangle->vindices[2]) + 1];
            *TmpVertexArrayPtr->Triangles++ = model->vertices[(3 * triangle->vindices[2]) + 2];
        }
        group = group->next;
    }
}

GLvoid __stdcall
glmDestroyVertexArray(GLMArray *VertexArray)
{
    if(VertexArray->Triangles) free(VertexArray->Triangles);
    if(VertexArray->Normals) free(VertexArray->Normals);
    if(VertexArray->Colors) free(VertexArray->Colors);
    if(VertexArray->Textures) free(VertexArray->Textures);
}


/* glmList: Generates and returns a display list for the model using
 * the mode specified.
 *
 * model - initialized GLMmodel structure
 * mode  - a bitwise OR of values describing what is to be rendered.
 *             GLM_NONE     -  render with only vertices
 *             GLM_FLAT     -  render with facet normals
 *             GLM_SMOOTH   -  render with vertex normals
 *             GLM_TEXTURE  -  render with texture coords
 *             GLM_COLOR    -  render with colors (color material)
 *             GLM_MATERIAL -  render with materials
 *             GLM_COLOR and GLM_MATERIAL should not both be specified.  
 * GLM_FLAT and GLM_SMOOTH should not both be specified.  
 */
int __stdcall
glmList(GLMmodel* model, int mode)
{
    int list;
    
    list = glGenLists(1);
    glNewList(list, GL_COMPILE);
    glmDraw(model, mode);
    glEndList();
    
    return list;
}

/* glmWeld: eliminate (weld) vectors that are within an epsilon of
 * each other.
 *
 * model   - initialized GLMmodel structure
 * epsilon     - maximum difference between vertices
 *               ( 0.00001 is a good start for a unitized model)
 *
 */
GLvoid __stdcall
glmWeld(GLMmodel* model, float epsilon)
{
    float* vectors;
    float* copies;
    int   numvectors;
    int   i;
    
    /* vertices */
    numvectors = model->numvertices;
    vectors  = model->vertices;
    copies = glmWeldVectors(vectors, &numvectors, epsilon);
    
//#if 0
  //  printf("glmWeld(): %d redundant vertices.\n", 
    //    model->numvertices - numvectors - 1);
//#endif
    
    for (i = 0; i < model->numtriangles; i++) {
        T(i).vindices[0] = (int)vectors[3 * T(i).vindices[0] + 0];
        T(i).vindices[1] = (int)vectors[3 * T(i).vindices[1] + 0];
        T(i).vindices[2] = (int)vectors[3 * T(i).vindices[2] + 0];
    }
    
    /* free space for old vertices */
    free(vectors);
    
    /* allocate space for the new vertices */
    model->numvertices = numvectors;
    model->vertices = (float*)malloc(sizeof(float) * 
        3 * (model->numvertices + 1));
    
    /* copy the optimized vertices into the actual vertex list */
    for (i = 1; i <= model->numvertices; i++) {
        model->vertices[3 * i + 0] = copies[3 * i + 0];
        model->vertices[3 * i + 1] = copies[3 * i + 1];
        model->vertices[3 * i + 2] = copies[3 * i + 2];
    }
    
    free(copies);
}

/* glmReadPPM: read a PPM raw (type P6) file.  The PPM file has a header
 * that should look something like:
 *
 *    P6
 *    # comment
 *    width height max_value
 *    rgbrgbrgb...
 *
 * where "P6" is the magic cookie which identifies the file type and
 * should be the only characters on the first line followed by a
 * carriage return.  Any line starting with a # mark will be treated
 * as a comment and discarded.   After the magic cookie, three integer
 * values are expected: width, height of the image and the maximum
 * value for a pixel (max_value must be < 256 for PPM raw files).  The
 * data section consists of width*height rgb triplets (one byte each)
 * in binary format (i.e., such as that written with fwrite() or
 * equivalent).
 *
 * The rgb data is returned as an array of unsigned chars (packed
 * rgb).  The malloc()'d memory should be free()'d by the caller.  If
 * an error occurs, an error message is sent to stderr and NULL is
 * returned.
 *
 * filename   - name of the .ppm file.
 * width      - will contain the width of the image on return.
 * height     - will contain the height of the image on return.
 *
 */
GLubyte* __stdcall
glmReadPPM(char* filename, int* width, int* height)
{
    FILE* fp;
    int i, w, h, d;
    unsigned char* image;
    char head[70];          /* max line <= 70 in PPM (per spec). */
    
    fp = fopen(filename, "rb");
    if (!fp) {
        perror(filename);
        return NULL;
    }
    
    /* grab first two chars of the file and make sure that it has the
       correct magic cookie for a raw PPM file. */
    fgets(head, 70, fp);
    if (strncmp(head, "P6", 2)) {
        //fprintf(stderr, "%s: Not a raw PPM file\n", filename);
        return NULL;
    }
    
    /* grab the three elements in the header (width, height, maxval). */
    i = 0;
    while(i < 3) {
        fgets(head, 70, fp);
        if (head[0] == '#')     /* skip comments. */
            continue;
        if (i == 0)
            i += sscanf(head, "%d %d %d", &w, &h, &d);
        else if (i == 1)
            i += sscanf(head, "%d %d", &h, &d);
        else if (i == 2)
            i += sscanf(head, "%d", &d);
    }
    
    /* grab all the image data in one fell swoop. */
    image = (unsigned char*)malloc(sizeof(unsigned char)*w*h*3);
    fread(image, sizeof(unsigned char), w*h*3, fp);
    fclose(fp);
    
    *width = w;
    *height = h;
    return image;
}

#if 0
/* normals */
if (model->numnormals) {
    numvectors = model->numnormals;
    vectors  = model->normals;
    copies = glmOptimizeVectors(vectors, &numvectors);
    
//    printf("glmOptimize(): %d redundant normals.\n", 
  //      model->numnormals - numvectors);
    
    for (i = 0; i < model->numtriangles; i++) {
        T(i).nindices[0] = (int)vectors[3 * T(i).nindices[0] + 0];
        T(i).nindices[1] = (int)vectors[3 * T(i).nindices[1] + 0];
        T(i).nindices[2] = (int)vectors[3 * T(i).nindices[2] + 0];
    }
    
    /* free space for old normals */
    free(vectors);
    
    /* allocate space for the new normals */
    model->numnormals = numvectors;
    model->normals = (float*)malloc(sizeof(float) * 
        3 * (model->numnormals + 1));
    
    /* copy the optimized vertices into the actual vertex list */
    for (i = 1; i <= model->numnormals; i++) {
        model->normals[3 * i + 0] = copies[3 * i + 0];
        model->normals[3 * i + 1] = copies[3 * i + 1];
        model->normals[3 * i + 2] = copies[3 * i + 2];
    }
    
    free(copies);
}

/* texcoords */
if (model->numtexcoords) {
    numvectors = model->numtexcoords;
    vectors  = model->texcoords;
    copies = glmOptimizeVectors(vectors, &numvectors);
    
//    printf("glmOptimize(): %d redundant texcoords.\n", 
 //       model->numtexcoords - numvectors);
    
    for (i = 0; i < model->numtriangles; i++) {
        for (j = 0; j < 3; j++) {
            T(i).tindices[j] = (int)vectors[3 * T(i).tindices[j] + 0];
        }
    }
    
    /* free space for old texcoords */
    free(vectors);
    
    /* allocate space for the new texcoords */
    model->numtexcoords = numvectors;
    model->texcoords = (float*)malloc(sizeof(float) * 
        2 * (model->numtexcoords + 1));
    
    /* copy the optimized vertices into the actual vertex list */
    for (i = 1; i <= model->numtexcoords; i++) {
        model->texcoords[2 * i + 0] = copies[2 * i + 0];
        model->texcoords[2 * i + 1] = copies[2 * i + 1];
    }
    
    free(copies);
}
#endif

#if 0
/* look for unused vertices */
/* look for unused normals */
/* look for unused texcoords */
for (i = 1; i <= model->numvertices; i++) {
    for (j = 0; j < model->numtriangles; i++) {
        if (T(j).vindices[0] == i || 
            T(j).vindices[1] == i || 
            T(j).vindices[1] == i)
            break;
    }
}
#endif
